import datetime
from flask import request, redirect

from flaskext.admin import Admin, ModelAdmin, AdminPanel

from app import app, db
from auth import auth
from models import User, Message, Note, Relationship


class NotePanel(AdminPanel):
    template_name = 'admin/notes.html'
    
    def get_urls(self):
        return (
            ('/create/', self.create),
        )
    
    def create(self):
        if request.method == 'POST':
            if request.form.get('message'):
                Note.create(
                    user=auth.get_logged_in_user(),
                    message=request.form['message'],
                )
        next = request.form.get('next') or self.dashboard_url()
        return redirect(next)
    
    def get_context(self):
        return {
            'note_list': Note.select().order_by(('created_date', 'desc')).paginate(1, 3)
        }

class UserStatsPanel(AdminPanel):
    template_name = 'admin/user_stats.html'
    
    def get_context(self):
        last_week = datetime.datetime.now() - datetime.timedelta(days=7)
        signups_this_week = User.filter(join_date__gt=last_week).count()
        messages_this_week = Message.filter(pub_date__gt=last_week).count()
        return {
            'signups': signups_this_week,
            'messages': messages_this_week,
        }


admin = Admin(app, auth)


class MessageAdmin(ModelAdmin):
    columns = ('user', 'content', 'pub_date',)

class NoteAdmin(ModelAdmin):
    columns = ('user', 'message', 'created_date',)


auth.register_admin(admin)
admin.register(Relationship)
admin.register(Message, MessageAdmin)
admin.register(Note, NoteAdmin)
admin.register_panel('Notes', NotePanel)
admin.register_panel('User stats', UserStatsPanel)



import math
import random
import re
import sys
from hashlib import sha1

from flask import abort, request, render_template
from peewee import Model, DoesNotExist, SelectQuery


def get_object_or_404(query_or_model, **query):
    try:
        return query_or_model.get(**query)
    except DoesNotExist:
        abort(404)

def object_list(template_name, qr, var_name='object_list', **kwargs):
    pq = PaginatedQuery(qr, kwargs.pop('paginate_by', 20))
    kwargs[var_name] = pq.get_list()
    return render_template(template_name, pagination=pq, page=pq.get_page(), **kwargs)


class PaginatedQuery(object):
    page_var = 'page'
    
    def __init__(self, query_or_model, paginate_by):
        self.paginate_by = paginate_by
        
        if isinstance(query_or_model, SelectQuery):
            self.query = query_or_model
            self.model = self.query.model
        else:
            self.model = query_or_model
            self.query = self.model.select()
    
    def get_page(self):
        return int(request.args.get(self.page_var) or 1)
    
    def get_pages(self):
        return math.ceil(float(self.query.count()) / self.paginate_by)
    
    def get_list(self):
        return self.query.paginate(self.get_page(), self.paginate_by)


def get_next():
    if not request.query_string:
        return request.path
    return '%s?%s' % (request.path, request.query_string)

def slugify(s):
    return re.sub('[^a-z0-9_\-]+', '-', s.lower())

def load_class(s):
    path, klass = s.rsplit('.', 1)
    __import__(path)
    mod = sys.modules[path]
    return getattr(mod, klass)


# borrowing these methods, slightly modified, from django.contrib.auth
def get_hexdigest(salt, raw_password):
    return sha1(salt + raw_password).hexdigest()

def make_password(raw_password):
    salt = get_hexdigest(str(random.random()), str(random.random()))[:5]
    hsh = get_hexdigest(salt, raw_password)
    return '%s$%s' % (salt, hsh)

def check_password(raw_password, enc_password):
    salt, hsh = enc_password.split('$', 1)
    return hsh == get_hexdigest(salt, raw_password)



import datetime

from flaskext.serializer import Serializer, ModelSerializer, Deserializer, ModelDeserializer
from flaskext.tests.base import FlaskPeeweeTestCase
from flaskext.tests.test_app import User, Message, Note


class SerializerTestCase(FlaskPeeweeTestCase):
    def setUp(self):
        super(SerializerTestCase, self).setUp()
        self.s = Serializer()
        self.ms = ModelSerializer()
        self.d = Deserializer()
        self.md = ModelDeserializer()
    
    def test_serializer(self):
        users = self.create_users()
        serialized = self.s.serialize_object(self.admin)
        self.assertEqual(serialized, {
            'id': self.admin.id,
            'username': 'admin',
            'password': self.admin.password,
            'join_date': self.admin.join_date.strftime('%Y-%m-%d %H:%M:%S'),
            'active': True,
            'admin': True,
            'email': '',
        })
        
        serialized = self.s.serialize_object(self.admin, fields=('id', 'username',))
        self.assertEqual(serialized, {
            'id': self.admin.id,
            'username': 'admin',
        })
        
        serialized = self.s.serialize_object(self.admin, exclude=('password', 'join_date',))
        self.assertEqual(serialized, {
            'id': self.admin.id,
            'username': 'admin',
            'active': True,
            'admin': True,
            'email': '',
        })
    
    def test_deserializer(self):
        users = self.create_users()
        
        deserialized = self.d.deserialize_object({
            'id': self.admin.id,
            'username': 'admin',
            'password': self.admin.password,
            'join_date': self.admin.join_date.strftime('%Y-%m-%d %H:%M:%S'),
            'active': True,
            'admin': True,
        }, User())
        
        for attr in ['id', 'username', 'password', 'active', 'admin']:
            self.assertEqual(
                getattr(deserialized, attr),
                getattr(self.admin, attr),
            )
        
        self.assertEqual(
            deserialized.join_date.strftime('%Y-%m-%d %H:%M:%S'),
            self.admin.join_date.strftime('%Y-%m-%d %H:%M:%S'),
        )
            
        admin_pk = self.admin.id
        
        deserialized = self.d.deserialize_object({
            'username': 'edited',
            'active': False,
            'admin': False,
        }, self.admin)
        
        self.assertEqual(deserialized.username, 'edited')
        self.assertEqual(deserialized.admin, False)
        self.assertEqual(deserialized.active, False)
        self.assertEqual(deserialized.id, admin_pk)
        
        deserialized.save()
        
        self.assertEqual(User.select().count(), 3)
        edited = User.get(username='edited')
        self.assertEqual(edited.id, admin_pk)
    
    def test_model_serializer(self):
        users = self.create_users()
        serialized = self.ms.serialize_object(self.admin)
        self.assertEqual(serialized, {
            '__model__': 'User',
            '__module__': 'flaskext.tests.test_app',
            'id': self.admin.id,
            'username': 'admin',
            'password': self.admin.password,
            'join_date': self.admin.join_date.strftime('%Y-%m-%d %H:%M:%S'),
            'active': True,
            'admin': True,
            'email': '',
        })
        
        serialized = self.ms.serialize_object(self.admin, fields=('id', 'username',))
        self.assertEqual(serialized, {
            '__model__': 'User',
            '__module__': 'flaskext.tests.test_app',
            'id': self.admin.id,
            'username': 'admin',
        })
    
    def test_model_deserializer(self):
        users = self.create_users()
        
        deserialized = self.md.deserialize_object({
            '__model__': 'User',
            '__module__': 'flaskext.tests.test_app',
            'id': self.admin.id,
            'username': 'admin',
            'password': self.admin.password,
            'join_date': self.admin.join_date.strftime('%Y-%m-%d %H:%M:%S'),
            'active': True,
            'admin': True,
        })
        
        for attr in ['id', 'username', 'password', 'active', 'admin']:
            self.assertEqual(
                getattr(deserialized, attr),
                getattr(self.admin, attr),
            )
        
        self.assertEqual(
            deserialized.join_date.strftime('%Y-%m-%d %H:%M:%S'),
            self.admin.join_date.strftime('%Y-%m-%d %H:%M:%S'),
        )
            
        admin_pk = self.admin.id
        
        deserialized = self.md.deserialize_object({
            '__model__': 'User',
            '__module__': 'flaskext.tests.test_app',
            'username': 'edited',
            'active': False,
            'admin': False,
        }, self.admin)
        
        self.assertEqual(deserialized.username, 'edited')
        self.assertEqual(deserialized.admin, False)
        self.assertEqual(deserialized.active, False)
        self.assertEqual(deserialized.id, admin_pk)
        
        deserialized.save()
        
        self.assertEqual(User.select().count(), 3)
        edited = User.get(username='edited')
        self.assertEqual(edited.id, admin_pk)
    
    def test_s_and_d(self):
        self.create_users()
        
        s = self.s.serialize_object(self.admin)
        d = self.d.deserialize_object(s, User())
        self.assertEqual(d, self.admin)
        
        s = self.ms.serialize_object(self.admin)
        d = self.md.deserialize_object(s)
        self.assertEqual(d, self.admin)



import datetime
import sys


class Serializer(object):
    date_format = '%Y-%m-%d'
    time_format = '%H:%M:%S'
    datetime_format = ' '.join([date_format, time_format])
    
    def convert_value(self, value):
        if isinstance(value, datetime.datetime):
            return value.strftime(self.datetime_format)
        elif isinstance(value, datetime.date):
            return value.strftime(self.date_format)
        elif isinstance(value, datetime.time):
            return value.strftime(self.time_format)
        else:
            return value
    
    def serialize_object(self, obj, fields=None, exclude=None):
        data = {}
        
        serialize_fields = []
        for f in obj._meta.get_field_names():
            if fields and f not in fields:
                continue
            if exclude and f in exclude:
                continue
            serialize_fields.append(f)
        
        for field_name in serialize_fields:
            data[field_name] = self.convert_value(getattr(obj, field_name))
        
        return data


class ModelSerializer(Serializer):
    def serialize_object(self, obj, fields=None, exclude=None):
        data = super(ModelSerializer, self).serialize_object(obj, fields, exclude)
        
        data['__module__'] = obj.__module__
        data['__model__'] = obj.__class__.__name__
        
        return data


class Deserializer(object):
    def deserialize_object(self, data, instance):
        for field, value in data.iteritems():
            if field in instance._meta.rel_fields:
                field = instance._meta.rel_fields[field]
            
            if field not in instance._meta.fields:
                continue
            
            field_obj = instance._meta.fields[field]
            
            setattr(instance, field, field_obj.python_value(value))
        
        return instance


class ModelDeserializer(Deserializer):
    def __init__(self):
        self._model_cache = {}
    
    def get_model(self, module, model):
        try:
            model_class = self._model_cache[(module, model)]
        except KeyError:
            try:
                __import__(module)
            except ImportError:
                raise ValueError('Unable to import module "%s"' % module)
            
            try:
                model_class = getattr(sys.modules[module], model)
            except AttributeError:
                raise ValueError('Model "%s" not found in module "%s"' % (model, module))
            
            self._model_cache[(module, model)] = model_class
        
        return model_class
    
    def deserialize_object(self, data, instance=None):
        if '__model__' not in data or '__module__' not in data:
            raise ValueError('Unable to deserialize, missing either "__model__" or "__module__" attribute')
        
        module, model = data.pop('__module__'), data.pop('__model__')
        if instance is None:
            instance = self.get_model(module, model)()
        
        return super(ModelDeserializer, self).deserialize_object(data, instance)



import functools
import operator
import os
import re

from flask import Blueprint, render_template, abort, request, url_for, redirect, flash
from flaskext.utils import get_next, PaginatedQuery, slugify
from peewee import BooleanField, TextField, Q
from wtforms import fields, widgets
from wtfpeewee.orm import model_form, ModelConverter


current_dir = os.path.dirname(__file__)


class BooleanSelectField(fields.SelectFieldBase):
    widget = widgets.Select()

    def iter_choices(self):
        yield ('1', 'True', self.data)
        yield ('', 'False', not self.data)

    def process_data(self, value):
        try:
            self.data = bool(value)
        except (ValueError, TypeError):
            self.data = None

    def process_formdata(self, valuelist):
        if valuelist:
            try:
                self.data = bool(valuelist[0])
            except ValueError:
                raise ValueError(self.gettext(u'Invalid Choice: could not coerce'))


def convert_boolean(model, field, **kwargs):
    return field.name, BooleanSelectField()

def convert_textfield(model, field, **kwargs):
    return field.name, fields.TextField()

converter = ModelConverter({
    BooleanField: convert_boolean,
})

filter_converter = ModelConverter({
    BooleanField: convert_boolean,
    TextField: convert_textfield,
})


class ModelAdmin(object):
    """
    ModelAdmin provides create/edit/delete functionality for a peewee Model.
    """
    paginate_by = 20
    columns = None
    ignore_filters = ('ordering', 'page',)
    
    def __init__(self, admin, model):
        self.admin = admin
        self.model = model
        self.pk_name = self.model._meta.pk_name
    
    def get_url_name(self, name):
        return '%s.%s_%s' % (
            self.admin.blueprint.name,
            self.get_admin_name(),
            name,
        )
    
    def get_form(self):
        return model_form(self.model, converter=converter)
    
    def get_filter_form(self):
        return model_form(self.model, converter=filter_converter)
    
    def get_add_form(self):
        return self.get_form()
    
    def get_edit_form(self):
        return self.get_form()
    
    def get_query(self):
        return self.model.select()
    
    def get_object(self, pk):
        return self.get_query().get(**{self.pk_name: pk})
    
    def get_urls(self):
        return (
            ('/', self.index),
            ('/add/', self.add),
            ('/delete/', self.delete),
            ('/<pk>/', self.edit),
        )
    
    def get_columns(self):
        return self.model._meta.get_field_names()
    
    def column_is_sortable(self, col):
        return col in self.model._meta.fields
    
    def get_display_name(self):
        return self.model.__name__
    
    def get_admin_name(self):
        return slugify(self.model.__name__)
    
    def save_model(self, instance, form, adding=False):
        form.populate_obj(instance)
        instance.save()
        return instance
    
    def index(self):
        query = self.get_query()
        form = self.get_filter_form()(request.args)
        
        ordering = request.args.get('ordering') or ''
        if ordering:
            desc, column = ordering.startswith('-'), ordering.lstrip('-')
            if self.column_is_sortable(column):
                query = query.order_by((column, desc and 'desc' or 'asc'))
        
        filters = []
        raw_filters = []
        for key in request.args:
            if key in self.ignore_filters:
                continue
            
            values = request.args.getlist(key)
            raw_filters.append((key, values))
            
            lookups = []
            
            for value in values:
                if value.startswith('*') and value.endswith('*'):
                    lookup = 'icontains'
                elif value.endswith('*'):
                    lookup = 'istartswith'
                else:
                    lookup = 'eq'
                
                value = value.strip('*')
                lookups.append(Q(**{'%s__%s' % (key, lookup): value}))
            
            if len(lookups) == 1:
                filters.append(lookups[0])
            else:
                filters.append(reduce(operator.or_, lookups))
        
        if filters:
            query = query.filter(*filters)
        
        pq = PaginatedQuery(query, self.paginate_by)
        
        if request.method == 'POST':
            id_list = request.form.getlist('id')
            return redirect(url_for(self.get_url_name('delete'), id=id_list))
        
        return render_template('admin/models/index.html',
            model_admin=self,
            query=pq,
            ordering=ordering,
            form=form,
            filters=filters,
            raw_filters=raw_filters,
        )
    
    def add(self):
        Form = self.get_add_form()
        
        if request.method == 'POST':
            form = Form(request.form)
            if form.validate():
                instance = self.save_model(self.model(), form, True)
                flash('New %s saved successfully' % self.get_display_name(), 'success')                
                return redirect(
                    url_for(self.get_url_name('edit'), pk=instance.get_pk())
                )
        else:
            form = Form()
        
        return render_template('admin/models/add.html', model_admin=self, form=form)
    
    def edit(self, pk):
        try:
            instance = self.get_object(pk)
        except self.model.DoesNotExist:
            abort(404)
        
        Form = self.get_edit_form()
        
        if request.method == 'POST':
            form = Form(request.form, obj=instance)
            if form.validate():
                self.save_model(instance, form, False)
                flash('Changes to %s saved successfully' % self.get_display_name(), 'success')
                return redirect(
                    url_for(self.get_url_name('edit'), pk=instance.get_pk())
                )
        else:
            form = Form(obj=instance)
        
        return render_template('admin/models/edit.html', model_admin=self, instance=instance, form=form)
    
    def delete(self):
        if request.method == 'GET':
            id_list = request.args.getlist('id')
            query = self.model.select().where(**{
                '%s__in' % self.model._meta.pk_name: id_list
            })
        elif request.method == 'POST':
            id_list = request.form.getlist('id')
            query = self.model.delete().where(**{
                '%s__in' % self.model._meta.pk_name: id_list
            })
            results = query.execute()
            flash('Successfully deleted %s %ss' % (results, self.get_display_name()), 'success')
            return redirect(url_for(self.get_url_name('index')))
        
        return render_template('admin/models/delete.html', model_admin=self, query=query)


class AdminPanel(object):
    template_name = 'admin/panels/default.html'
    
    def __init__(self, admin, title):
        self.admin = admin
        self.title = title
        self.slug = slugify(self.title)
    
    def dashboard_url(self):
        return url_for('%s.index' % (self.admin.blueprint.name))
    
    def get_urls(self):
        return ()
    
    def get_url_name(self, name):
        return '%s.panel_%s_%s' % (
            self.admin.blueprint.name,
            self.slug,
            name,
        )
    
    def get_template_name(self):
        return self.template_name
    
    def get_context(self):
        return {}
    
    def render(self):
        return render_template(self.get_template_name(), panel=self, **self.get_context())


class AdminTemplateHelper(object):
    def __init__(self, admin):
        self.admin = admin
        self.app = self.admin.app
    
    def get_model_field(self, model, field):
        attr = getattr(model, field)
        if callable(attr):
            return attr()
        return attr
    
    def get_form_field(self, form, field_name):
        return getattr(form, field_name)
    
    def fix_underscores(self, s):
        return s.replace('_', ' ').title()
    
    def update_querystring(self, querystring, key, val):
        if not querystring:
            return '%s=%s' % (key, val)
        else:
            querystring = re.sub('%s(?:[^&]+)?&?' % key, '', querystring).rstrip('&')
            return ('%s&%s=%s' % (querystring, key, val)).lstrip('&')
    
    def get_verbose_name(self, model, column_name):
        try:
            field = model._meta.fields[column_name]
        except KeyError:
            return self.fix_underscores(column_name)
        else:
            return field.verbose_name
    
    def get_model_admins(self):
        return {'model_admins': self.admin.get_model_admins()}
    
    def prepare_environment(self):
        self.app.template_context_processors[None].append(self.get_model_admins)
        
        self.app.jinja_env.globals['get_model_field'] = self.get_model_field
        self.app.jinja_env.globals['get_form_field'] = self.get_form_field
        self.app.jinja_env.globals['get_verbose_name'] = self.get_verbose_name
        self.app.jinja_env.filters['fix_underscores'] = self.fix_underscores
        self.app.jinja_env.globals['update_querystring'] = self.update_querystring


class Admin(object):
    def __init__(self, app, auth, template_helper=AdminTemplateHelper,
                 prefix='/admin', name='admin'):
        self.app = app
        self.auth = auth
        
        self._registry = {}
        self._panels = {}
        
        self.blueprint = self.get_blueprint(name)
        self.url_prefix = prefix
        
        self.template_helper = template_helper(self)
        self.template_helper.prepare_environment()
    
    def auth_required(self, func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            user = self.auth.get_logged_in_user()
            
            if not user:
                login_url = url_for('%s.login' % self.auth.blueprint.name, next=get_next())
                return redirect(login_url)
            
            if not self.check_user_permission(user):
                abort(403)
            
            return func(*args, **kwargs)
        return inner
    
    def check_user_permission(self, user):
        return user.admin
    
    def get_urls(self):
        return (
            ('/', self.auth_required(self.index)),
        )
    
    def register(self, model, admin_class=ModelAdmin):
        model_admin = admin_class(self, model)
        admin_name = model_admin.get_admin_name()
        
        self._registry[admin_name] = model_admin
    
    def unregister(self, model):
        del(self._registry[model])
    
    def register_panel(self, title, panel):
        panel_instance = panel(self, title)
        self._panels[title] = panel_instance
    
    def unregister_panel(self, title):
        del(self._panels[title])
    
    def get_model_admins(self):
        return sorted(self._registry.values(), key=lambda o: o.get_admin_name())
    
    def get_panels(self):
        return sorted(self._panels.values(), key=lambda o: o.slug)
    
    def index(self):
        return render_template('admin/index.html',
            model_admins=self.get_model_admins(),
            panels=self.get_panels(),
        )
    
    def get_blueprint(self, blueprint_name):
        return Blueprint(
            blueprint_name,
            __name__,
            static_folder=os.path.join(current_dir, 'static'),
            template_folder=os.path.join(current_dir, 'templates'),
        )
    
    def register_blueprint(self, **kwargs):
        self.app.register_blueprint(
            self.blueprint,
            url_prefix=self.url_prefix,
            **kwargs
        )
    
    def configure_routes(self):
        for url, callback in self.get_urls():
            self.blueprint.route(url, methods=['GET', 'POST'])(callback)
        
        for model_admin in self._registry.values():
            admin_name = model_admin.get_admin_name()
            for url, callback in model_admin.get_urls():  
                full_url = '/%s%s' % (admin_name, url)
                self.blueprint.add_url_rule(
                    full_url,
                    '%s_%s' % (admin_name, callback.__name__),
                    self.auth_required(callback),
                    methods=['GET', 'POST'],
                )
        
        for panel in self._panels.values():
            for url, callback in panel.get_urls():
                full_url = '/%s%s' % (panel.slug, url)
                self.blueprint.add_url_rule(
                    full_url,
                    'panel_%s_%s' % (panel.slug, callback.__name__),
                    self.auth_required(callback),
                    methods=['GET', 'POST'],
                )
    
    def setup(self):
        self.configure_routes()
        self.register_blueprint()



import datetime

from flask import request, session, url_for, get_flashed_messages

from flaskext.auth import Auth, LoginForm
from flaskext.tests.base import FlaskPeeweeTestCase
from flaskext.tests.test_app import User, app, db, auth


class TestAuth(Auth):
    def setup(self):
        pass


class AuthTestCase(FlaskPeeweeTestCase):
    def setUp(self):
        super(AuthTestCase, self).setUp()
        
        self.test_auth = TestAuth(app, db)
    
    def login(self, username='admin', password='admin', context=None):
        context = context or self.app
        return context.post('/accounts/login/', data={
            'username': username,
            'password': password,
        })
    
    def logout(self, context=None):
        context = context or self.app
        return context.post('/accounts/logout/')
    
    def test_login_view(self):
        self.create_users()
        
        with self.flask_app.test_client() as c:
            resp = c.get('/accounts/login/')
            self.assertEqual(resp.status_code, 200)
            
            # check that we have no logged-in user
            self.assertContext('user', None)
            
            frm = self.get_context('form')
            self.assertTrue(isinstance(frm, LoginForm))
            self.assertEqual(frm.data, {'username': None, 'password': None})
            
            # make a post missing the username
            resp = c.post('/accounts/login/', data={
                'username': '',
                'password': 'xxx',
            })
            self.assertEqual(resp.status_code, 200)
            
            # check form for errors
            frm = self.get_context('form')
            self.assertEqual(frm.errors, {'username': [u'This field is required.']})
            
            # check that no messages were generated
            self.assertFalse('_flashes' in session)
            
            # check that the auth API does not indicate a logged-in user
            self.assertEqual(auth.get_logged_in_user(), None)
            
            # make a post with a bad username/password combo
            resp = c.post('/accounts/login/', data={
                'username': 'normal',
                'password': 'baz',
            })
            self.assertEqual(resp.status_code, 200)
            
            # both fields were present so no form errors, but flash the user
            # indicating bad username/password combo
            self.assertTrue('_flashes' in session)
            messages = get_flashed_messages()
            
            self.assertEqual(messages, [
                'Incorrect username or password',
            ])
            
            # check that the auth API does not indicate a logged-in user
            self.assertEqual(auth.get_logged_in_user(), None)

            # make a post with an inactive user
            resp = c.post('/accounts/login/', data={
                'username': 'inactive',
                'password': 'inactive',
            })
            self.assertEqual(resp.status_code, 200)
            
            # still no logged-in user
            self.assertContext('user', None)
            
            # check that the auth API does not indicate a logged-in user
            self.assertEqual(auth.get_logged_in_user(), None)
            
            # finally post as a known good user
            resp = c.post('/accounts/login/', data={
                'username': 'normal',
                'password': 'normal',
            })
            self.assertEqual(resp.status_code, 302)
            
            # check that we now have a logged-in user
            self.assertEqual(auth.get_logged_in_user(), self.normal)
    
    def test_login_redirect(self):
        self.create_users()
        
        with self.flask_app.test_client() as c:
            resp = c.post('/accounts/login/?next=/foo-baz/', data={
                'username': 'normal',
                'password': 'normal',
            })
            self.assertEqual(resp.status_code, 302)
            self.assertTrue(resp.headers['location'].endswith('/foo-baz/'))
    
    def test_login_logout(self):
        self.create_users()
        
        with self.flask_app.test_client() as c:
            resp = c.post('/accounts/login/', data={
                'username': 'normal',
                'password': 'normal',
            })
            self.assertEqual(auth.get_logged_in_user(), self.normal)
            
            resp = c.post('/accounts/logout/')
            self.assertEqual(auth.get_logged_in_user(), None)
            
            resp = c.post('/accounts/login/', data={
                'username': 'admin',
                'password': 'admin',
            })
            self.assertEqual(auth.get_logged_in_user(), self.admin)
            
            # log back in without logging out
            resp = c.post('/accounts/login/', data={
                'username': 'normal',
                'password': 'normal',
            })
            self.assertEqual(auth.get_logged_in_user(), self.normal)
    
    def test_login_required(self):
        self.create_users()
        
        with self.flask_app.test_client() as c:
            resp = c.get('/private/')
            self.assertEqual(resp.status_code, 302)
            self.assertTrue(resp.headers['location'].endswith('/accounts/login/?next=%2Fprivate%2F'))
            
            self.login('normal', 'normal', c)
            
            resp = c.get('/private/')
            self.assertEqual(resp.status_code, 200)

            self.assertEqual(auth.get_logged_in_user(), self.normal)
            
            self.login('admin', 'admin', c)
            
            resp = c.get('/private/')
            self.assertEqual(resp.status_code, 200)
            
            self.assertEqual(auth.get_logged_in_user(), self.admin)
